package com.recordingblogs.exampledelay;

import java.io.*;
import java.util.Vector;
import com.recordingblogs.oreffect.*;

public class Delay extends Undo implements EffectInterface
{
	/**
	 * The minimum allowed delay in seconds
	 */
	public static final float MINDELAY = 0;

	/**
	 * The maximum allowed delay in seconds
	 */
	public static final float MAXDELAY = 1;
	
	/**
	 * The minimum allowed decay ratio (i.e., 0.8 means 80% decay and an 
	 * output signal that has 80% of the amplitude of the input signal)
	 */
	public static final float MINDECAY = 0;

	/**
	 * The maximum allowed decay ratio (i.e., 0.8 means 80% decay and an 
	 * output signal that has 80% of the amplitude of the input signal)
	 */
	public static final float MAXDECAY = 1;
	
    /**
     * The amount of delay in seconds
     */
    private float m_delay;
	
    /**
     * The amount of decay
     */
    private float m_decay;
	
    /**
     * Store the incoming buffers, since a delay effect will have to look at previous
     * buffers to compute delayed values.  This array works as a FIFO queue, where 
     * once past buffers become useless (they are too old to be ever accessed given 
     * the maximum amount of delay), they are dropped
     */
    private transient Vector<int []> m_storeBuffers;

    /**
     * The original value of the delay is stored, as we are working with 
     * sliders.  We want the sliders to modify the amount of delay every time they move,
     * but we want undo information to only be stored once the focus of the slider is lost,
     * so that we do not store too many undos.  Thus, while the slider changes {@link #m_delay},
     * this variable preserves the original value of the delay before working with the
     * specific slider
     */
    private transient float m_delayOriginal;
    
    /**
     * The original value of the decay.  See {@link #m_delayOriginal} for
     * more information.
     */
    private transient float m_decayOriginal;
    
    /**
     * The default and only constructor
     */
    public Delay()
    {
        m_delay = 0.5F;
        m_decay = 0.5F;
        
        m_storeBuffers = new Vector<>(0);
        
        m_delayOriginal = m_delay;
        m_decayOriginal = m_decay;
    }

    /**
     * Get the value of the delay
     * @return A float between {@link #MINDELAY} and {@link #MAXDELAY}. The value of the 
     *         delay
     */
    public float getDelay()
    {
        return m_delay;
    }

    /**
     * Set the value of the delay ({@link #m_delay}).
     * @param delay A float between {@link #MINDELAY} and {@link #MAXDELAY}. The value of the 
     *        delay
     * @param storeUndo true, if this function should store undo information; false
     *        otherwise
     */
    public void setDelay(float delay, boolean storeUndo)
    {
    	if (delay < MINDELAY || delay > MAXDELAY)
    		System.out.println("Error in Delay::setDelay: Error in delay");	//$NON-NLS-1$
    	delay = Math.min(MAXDELAY, Math.max(MINDELAY, delay));
        m_delay = delay;
        
    	if (m_delayOriginal != delay && storeUndo)
    	{
    		m_delay = m_delayOriginal;
//    		fireEffectUndoEvent(new EffectUndoEvent(this, "Undo Delay"));
            m_delay = delay;
    		m_delayOriginal = m_delay;
    	}
    }

    /**
     * Get the value of the decay
     * @return A float between {@link #MINDECAY} and {@link #MAXDECAY}. The value of the 
     *         decay
     */
    public float getDecay()
    {
        return m_decay;
    }

    /**
     * Set the value of the decay ({@link #m_decay}).
     * @param decay A float between {@link #MINDECAY} and {@link #MAXDECAY}. The value of the 
     *        delay
     * @param storeUndo true, if this function should store undo information; false
     *        otherwise
     */
    public void setDecay(float decay, boolean storeUndo)
    {
    	if (decay < MINDECAY || decay > MAXDECAY)
    		System.out.println("Error in Delay::setDecay: Error in decay");	//$NON-NLS-1$
    	decay = Math.min(MAXDECAY, Math.max(MINDECAY, decay));
        m_decay = decay;
        
    	if (m_decayOriginal != decay && storeUndo)
    	{
    		m_decay = m_decayOriginal;
  //  		fireEffectUndoEvent(new EffectUndoEvent(this, "Undo Decay"));
            m_decay = decay;
    		m_decayOriginal = m_decay;
    	}
    }

    @Override
    public boolean allowsDryWetMix()
    {
    	return true;
    }
    
    @Override
    public boolean allowsSideChaining()
    {
    	return false;
    }

    @Override
    public boolean startPlay()
    {
        m_storeBuffers.setSize(0);
        return true;
    }

    /**
     * Perform any actions that should be performed after playback stops.
     * It is dangerous to remove stored audio data here, as access to these
     * data may still be needed shortly after the application is told to stop
     * playing.  Removing stored audio data is thus done at the beginning of 
     * the next playback
     */
    @Override
    public void stopPlay()
    {
    	// nothing to do
    }
    
    @Override
    public boolean hasData()
    {
    	if (m_storeBuffers.size() > 0)
    		return true;
    	return false;
    }

    @Override
    public void setLanguage(String language)
    {
    	// nothing to do; not implemented
    }

    @Override
    public boolean writeObject(WriteInterface ar)
    {
    	try
    	{
    		ar.writeFloat(3.0F);		// version number
	        ar.writeFloat(m_delay);
	        ar.writeFloat(m_decay);
    	}
    	catch (IOException e)
    	{
    		System.out.println("Exception in Delay::writeObject: " + e);	//$NON-NLS-1$
    		return false;
    	}
    	
    	return true;
    }

    @Override
    public boolean readObject(ReadInterface ar)
    {
    	try
    	{
    		ar.readFloat();		// version number, do something with it, if needed
	        m_delay = ar.readFloat();
	        m_decay = ar.readFloat();
    	}
    	catch (IOException e)
    	{
    		System.out.println("Exception in Delay::readObject: " + e);	//$NON-NLS-1$
    		return false;
    	}
    	
        m_delayOriginal = m_delay;
        m_decayOriginal = m_decay;
        
        return true;
    }

    @Override
    public void setEqual(EffectInterface effect)
    {
    	if (this == effect)
    		return;
    	if (effect.getClass() != this.getClass())
    	{
    		System.out.println("Error in Delay::setEqual: Wrong class");	//$NON-NLS-1$
    		return;
    	}
    	Delay e = (Delay) effect;
        m_delay = e.m_delay;
        m_decay = e.m_decay;

        m_delayOriginal = m_delay;
        m_decayOriginal = m_decay;
    }

	/**
	 * Apply the effect taking a "dry" buffer and returning a "wet" buffer.  This
	 * function, as implemented below, is only an example and only works on 16-bit 
	 * data.  Additional work is needed to make the function work on 8, 24, and 32 bit data.
	 * 
	 * The delay effect implemented here uses a single amount of delay and a single 
	 * amount of decay.  A more practical delay, such as the Orinj Delay, will allow
	 * for different delay and decay amounts in the left and right channel, which 
	 * could create a more interesting effect.
	 * 
	 * Still other delay effects may use multiple repetitions (the Orinj Echo and
	 * the Orinj Multitap Delay), delay sweeps (the Orinj Chorus and the Orinj Bass
	 * Chorus), or decay sweeps.
	 * 
	 * @param buffer  	The incoming sound data buffer to which the effect should be applied
	 *                  and the resulting outgoing sound data buffer with the effect applied
	 *
	 *                  This is an array of integers with values between Integer.MIN_VALUE
	 *                  and Integer.MAX_VALUE, such as for 32-bit integer PCM sound data.
	 *                  This is so for any sampling resolution that may be used in the rest of the
	 *                  application and applies to both the incoming data and the outgoing
     *                  data.
	 * @param controlbuffer	The control buffer should be used by effects that allow
	 * 					side chaining, such as a side chained compressor.  A side chained
	 * 					compressor changes the dynamics of one track based on the dynamics of
	 * 					a second track.  The audio data contained in the second track would
	 * 					be sent to the effect in this parameter.  An effect that uses this
	 * 					parameter should also return true in allowsSideChaining.  If the effect
	 * 					does not allow side chaining, this parameter may be null and should
	 * 					not be used.  The Orinj application controls which track is to be used
	 * 					as a control track.
     * @param channels  The number of audio channels in buffer.
     * @param channel   The channel to which this effect applies (see EffectChannels.java)
	 * @param samplingRate The sampling rate of the audio in buffer.
     * @param time      The start time in seconds of the audio buffers (e.g., buffer) in the
     *                  session or wave that is being played.  For example, if playback starts at
     *                  10 seconds into the multitrack session, time will be equal to 10 in the
     *                  first call to this function.
     * @param drymix    The dry mix of the effect.  This is a number between 0 and 1.  It is the
     *                  gain applied to the original signal before that original signal is placed
     *                  in buffer.  Effects that do not use dry or wet mix should not use this
     *                  argument.
     *
     *                  Not all effects have a dry and wet mix.  Consider a simple delay that
     *                  produces one delayed and decayed repetition of the original signal.
     *                  Implementing dry and wet mix in this effect is simple.  The dry mix gain
     *                  is applied to the original signal and the wet mix gain is applied
     *                  to the delayed and decayed repetition.  Typically, the dry mix is only
     *                  applied to the original signal after the delayed and decayed repetition
     *                  is computed, so that the dry mix does not impact the wet portion of
     *                  the resulting signal.  Similarly, the wet mix should not impact the
     *                  dry, original portion of the signal.
     *
     *                  A compressor, on the other hand, will not typically have dry or wet mix.
     *                  In a compressor, the arguments drymix, wetmix, dryMixEnvelope, and
     *                  wetMixEnvelope can be ignored.
     * @param wetmix    The wet mix of the effect.  This is a number between 0 and 1.  It is the
     *                  gain applied to the processed signal after that processed signal is
     *                  computed from the original signal.  Effects that do not use dry or wet
     *                  mix should not use this argument.
     * @param dryMixEnvelope    A gain envelope for the dry mix of the effect.  This envelope
     *                  is an additional dry mix gain to be applied to the original signal
     *                  before this original signal is placed in buffer and after this original
     *                  signal is used to compute the wet signal.  Effects that do not have dry
     *                  and wet mix should not use this argument.
     *
     *                  The difference between this argument and drymix is that drymix is constant
     *                  throughout the audio data in buffer and the dry mix envelope can change.
     *                  A call to the member function of this argument Envelope::getValueAt
     *                  will return the gain to be applied at a specific time.  The gain is
     *                  between 0 and 1 and should be used in the same way as drymix.
     * @param wetMixEnvelope    A gain envelope for the wet mix of the effect.  This envelope
     *                  is an additional wet mix gain to be applied to the processed signal
     *                  after this processed signal is computed from the original signal.
     *                  Effects that do not have dry and wet mix should not use this argument.
     * @param timeStart The beginning of when the effect should be applied, in seconds, from the
     * 					beginning of the session or wave.  Effects should be expect that this 
     * 					function will not be called if the sound buffer is outside of the interval
     * 					between timeStart and timeEnd
     * @param timeEnd	The end of when the effect should be applied, in seconds, from the beginning
     * 					of the session or wave.  Effects should be expect that this 
     * 					function will not be called if the sound buffer is outside of the interval
     * 					between timeStart and timeEnd
	 */
    @Override
    public boolean apply(int [] buffer, int [] controlbuffer, int channels, int channel, float samplingRate, double time, float drymix, float wetmix, 
    		EnvelopeInterface dryMixEnvelope, EnvelopeInterface wetMixEnvelope, double timeStart, double timeEnd)
    {
    	// if the all of buffer is before timeStart, nothing to do
    	if (timeStart >= 0 && time + (double) (buffer.length / channels) / samplingRate < timeStart)
    		return true;
    	
    	// if the all of buffer is after timeEnd and the delay, nothing to do
    	if (timeEnd >= 0 && time > timeEnd + m_delay)
    		return true;
    	
        // store the incoming audio data, which may be used during this call
        // to the function, but may also be used during subsequent calls to this
        // function 
        int [] storeBuffer = new int [buffer.length];
        System.arraycopy(buffer, 0, storeBuffer, 0, buffer.length);
        m_storeBuffers.add(storeBuffer);

        // the values of the wet buffer are the delayed repetitions of the signal
        // Thus, the current value of the wet buffer is equal to some past
        // value in the incoming data (with some decay).  Compute here where
        // the first value of wet buffer should be taken from (how far back
        // the delay needs to look.  We will start with the first available 
        // buffer in m_storeBuffers and will compute the actual byte position
        // from the amount of delay
        int curBuffer = m_storeBuffers.size() - 1;
        int curByte = (int) (m_delay * samplingRate * channels);
        while (curByte > 0 && curByte % channels != 0)
            curByte--;
        curByte = Math.max(curByte, 0);
        curByte = -curByte;
        while (curByte < 0)
        {
            curByte += buffer.length;
            curBuffer--;
        }
        int bufferFrom[] = null;
        if(curBuffer >= 0)
            bufferFrom = m_storeBuffers.get(curBuffer);

        // if we do not have to look too far back into the stored audio buffers,
        // drop the ones that we will not use
        while (curBuffer > 0)
        {
            m_storeBuffers.remove(0);
            curBuffer--;
        }

        // The dry and wet mix envelopes allow changes in the values of dry
        // and wet mix during playback.  We use these variables to get the
        // dry and wet mix values at specific playback times
        float drymixenvelope = 0;
        float wetmixenvelope = 0;
        double currentTime = time;
        
        // loop through the length of the audio buffer and compute its values
        // with the effect.  The implementation below will work for mono audio
        // (one channel) and for stereo audio (two channels).  
        //
        // if the left and right channel in stereo data has different delay and decay 
        // parameters, the values of the audio in the two channels should be computed 
        // separately.  If an effect computes the left and right channels separately, 
        // mono audio should be assumed to have a "left" channel only.
        for(int i = 0; i < buffer.length; curByte++, i++)
        {
        	// compute the time of the current sample and the values of the dry and 
        	// wet mix envelopes at that time. It is not necessary to recompute
        	// the dry and wet mix envelope values for every sample, as the human ear 
        	// is not that sensitive.  We could, for example, do this for every 50 
        	// miliseconds to reduce the number of computations
        	currentTime = time + ((double) i / channels) / samplingRate; 
        	drymixenvelope = (float) dryMixEnvelope.getValueAt(currentTime);
        	wetmixenvelope = (float) wetMixEnvelope.getValueAt(currentTime);
        	
        	// if curBuffer is negative, we are looking too far in the past
        	// and we do not have enough stored audio information.  This will
        	// happen at the beginning of playback.  Then, instead of the delay,
        	// the wet buffer will be filled with silence as below
        	if (curBuffer >= 0 && bufferFrom != null)
        	{
            	// adjust the current index of the stored buffer, if needed
	            if(curByte >= bufferFrom.length)
	            {
	                curBuffer++;
	                bufferFrom = m_storeBuffers.get(curBuffer);
	                curByte = 0;
	            }
	            
	            // compute the delay and apply the dry and wet mix and the dry and
	            // wet mix envelopes.  Place the sum of the original signal and the 
	            // wet signal (the delay repetition) in the buffer, since the same
	            // buffer carries both the input data and the output data
                buffer[i] = (int) Math.min(Math.max(buffer[i] * drymix * drymixenvelope + bufferFrom[curByte] * m_decay * wetmix * wetmixenvelope, Integer.MIN_VALUE), Integer.MAX_VALUE);
        	}
        }
        
        return true;
    }
}
